# Django settings for foo project.

DEBUG = True
TEMPLATE_DEBUG = DEBUG

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3', # Add 'postgresql_psycopg2', 'postgresql', 'mysql', 'sqlite3' or 'oracle'.
        'NAME': 'dfk.db',                      # Or path to database file if using sqlite3.
    }
}

ROOT_URLCONF = 'test_urls'

INSTALLED_APPS = (
    'dfk',
)



from django.db.models import ForeignKey
from django.db.models.loading import get_app
from django.db.models.loading import get_models
from dfk.models import DeferredForeignKey

def point(from_model, rel_name, to_model, **fk_kwargs):
    """
    Cause the foreign key `rel_name` on `from_model` to point to
    `to_model`. `fk_kwargs` will be forwarded to the generated
    foreign key.
    """
    assert not to_model._meta.abstract
    deferred_fk = getattr(from_model, rel_name)
    if not isinstance(deferred_fk, DeferredForeignKey):
        raise ValueError, u'You only point a DeferredForeignKey'

    target_model = '%s.%s' % (
        to_model._meta.app_label,
        to_model._meta.object_name
    )

    new_kwargs = deferred_fk.kwargs.copy()
    new_kwargs.update(fk_kwargs)
    new_foreign_key = ForeignKey(
        target_model,
        *deferred_fk.args,
        **new_kwargs
    )
    from_model.add_to_class(rel_name, new_foreign_key)

def point_named(app_name, target_name, to_model, **fk_kwargs):
    # Grab all the models from the app
    if isinstance(app_name, basestring):
        app = get_app(app_name)
    else:
        app = app_name
    models = get_models(app, include_deferred=True)

    # Now look for dfk instances with the target name
    for model in models:
        for attr in dir(model):
            ob = getattr(model, attr)
            if not isinstance(ob, DeferredForeignKey):
                continue

            # We found a deferred foreign key. Check the name
            # to see if it's the one we're processing.
            if getattr(ob, 'name', None) != target_name:
                continue

            # We found a foreign key. Repoint it.
            point(model, attr, to_model, **fk_kwargs)


def repoint(from_model, rel_name, to_model, **kwargs):
    options = getattr(to_model, '_meta', None)
    if options:
        assert not options.abstract, u'Target model may not be abstract'
    field = from_model._meta.get_field(rel_name)
    rel = field.rel
    args = (to_model, rel.field_name)
    new_kwargs = {
        'related_name': rel.related_name,
        'limit_choices_to': rel.limit_choices_to,
        'lookup_overrides': rel.lookup_overrides,
        'parent_link': rel.parent_link
    }
    new_kwargs.update(kwargs)
    new_rel = rel.__class__(*args, **new_kwargs)
    field.rel = new_rel



from django.db import models
from django.test import TestCase

from dfk import point
from dfk import point_named
from dfk import repoint
from dfk.models import DeferredForeignKey

class ModelA(models.Model):
    pass

class ModelB(models.Model):
    fk = models.ForeignKey(ModelA)

class ModelC(models.Model):
    pass

class ModelD(models.Model):
    fk = DeferredForeignKey(name='content')

class NotAModel(object):
    pass

class AbstractModel(models.Model):
    class Meta:
        abstract = True

# Repoint ModelB's foreign key field 'fk' to ModelC
repoint(ModelB, 'fk', ModelC)

class ExistingTestCase(TestCase):

    def test_existing_fk(self):
        # We've repointed ModelB's 'fk' field to reference
        # ModelC, rather than ModelA. Check this has worked
        # as expected
        c = ModelC.objects.create()
        b = ModelB.objects.create(fk=c)
        b.save()

        x = ModelB.objects.get(pk=b.pk)
        self.assertEqual(c, b.fk)

    def test_not_a_model(self):
        self.assertRaises(AssertionError, repoint, ModelB, 'fk', NotAModel)

    def test_abstract(self):
        self.assertRaises(AssertionError, repoint, ModelB, 'fk', AbstractModel)


class DeferredModelA(models.Model):
    user = DeferredForeignKey()
    
class DeferredModelB(models.Model):
    user = DeferredForeignKey()

# Point DeferredModelA's 'user' dfk to ModelA
point(DeferredModelA, 'user', ModelA, related_name='foo_set')

# Point ModelD's deferred fk to ModelA
point_named('dfk', 'content', ModelA)

class DeferredForeignKeyTestCase(TestCase):

    def test_specific_field(self):
        # Check that DeferredModelA's deferred field was correctly
        # processed
        m = ModelA.objects.create()
        a = DeferredModelA.objects.create(
            user=m
        )
        x = DeferredModelA.objects.get(pk=a.pk).user
        self.assertEqual(m, x)
        
        # Also test that the related_name override was passed
        # through
        self.assertEqual(1, len(x.foo_set.all()))

    def test_duplicate(self):
        # Attempting to point a dfk to a different target once
        # it's already been set should raise an error
        self.assertRaises(ValueError, point, DeferredModelA, 'user', ModelC)
        
    def test_point_non_deferred(self):
        # Attempting to point a non-DeferredForeignKey should raise a 
        # ValueError
        self.assertRaises(ValueError, point, ModelB, 'fk', ModelC)
        
    def test_point_abstract(self):
        # Attempting to point a dfk to an abstract class should raise
        # a AssertionError
        self.assertRaises(AssertionError, point, DeferredModelB, 'user', AbstractModel)
        
    def test_named_deferred(self):
        # Check that our repoint of the named foreign key worked
        a = ModelA.objects.create()
        d = ModelD.objects.create(fk=a)
        self.assertEqual(a, ModelD.objects.get(pk=d.pk).fk)
        


import os
from setuptools import setup, find_packages

version = '0.0.1'

setup(
    name='django-dfk',
    description="Deferred foreign keys for Django",
    version=version,
    long_description=open("README.rst").read() + "\n" +
    	             open(os.path.join("docs", "HISTORY.rst")).read(),
    author='Dan Fairs',
    author_email='dan@fezconsulting.com',
    license='BSD',
    packages=find_packages(exclude=['ez_setup']),
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Framework :: Django",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: BSD License"
        ],
)
